(*
KOL2FPC (C) by Vladimir Kladov, 2003
VERSION 1.6, 30-May-2003, upd: 5-Jun
------------------------------------
This is a utility to convert KOL.PAS
created initially for Delphi2-7 only
to \FPC\KOL.PAS which is compatible
with Free Pascal Compiler and can be
used with existing KOL and / or MCK 
projects to compile / build it using
FPC.

The source of KOL2FPC is provided.

KOL.PAS of version 1.75 or higher
required, containing directives
{#...}.

====================================
Usage:unpack KOL2FPC.EXE to the same
directory where KOL.PAS is located,
and run it there (double-click in
the Explorer window, for example).

A list of directives (may be yet not
complete), not case-sensitive:

{-} this text is removed {+}
{++}(|* this text is inserted*|){--}
{#Replace[text1][text2]}
{#Noreplace[text]}
{#SkipBetween[text1][text2]}
{#NoSkipBetween[text1][text2]}
{#Line} -> inserts {KOL.PAS Line #}
{#Include[path]} include as is, i.e.
              no directives there
{#If[symbol]} skip all the KOL2FPC
              commands in KOL.PAS
              until {#End} if symbol
              is defined
{#IfNot[symbol]} or {#If not[symbol}
(symbols are defined in command line
using key /Dsymbol or /D:symbol )
{#End} or {#EndIf}
 or even {#EndXXXXXXXXXXX}
{#List} writes current settings in
	(|* *|)-style comments

Other possible directives (may be
added in future versions):
{#Output[path]}
{#Else} or {#Else[symbol]}

====================================
http: //bonanzas.rinet.ru
mailto: bonanzas@online.sinor.ru
*)

program KOL2FPC;

{$APPTYPE CONSOLE}

uses
  Windows, KOL;

function StrLEq( S1, S2: PChar; L: Integer ): Boolean;
asm
  PUSH ESI
  MOV  ESI, EAX
  CMP  ESI, EDX
  JZ   @@exit  // Z -> TRUE
  JECXZ @@exit // NZ -> FALSE
@@loop:
  LODSB
  MOV AH, AL
  OR  AH, [EDX]
  JZ  @@exit
  MOV AH, [EDX]
  INC EDX

  SUB AL, 'A'
  CMP AL, 'Z'-'A'
  JA  @@1
  ADD AL, $20
@@1:
  SUB AH, 'A'
  CMP AH, 'Z'-'A'
  JA  @@2
  ADD AH, $20
@@2:
  CMP AL, AH
  JNZ @@exit  // NZ -> FALSE
  LOOP @@loop
@@exit:
  SETZ AL
  POP  ESI
end;

{function StrLEq( S1, S2: PChar; L: Integer ): Boolean;
var A1, A2: String;
begin
  A1 := S1;
  A1 := Copy( A1, 1, L );
  A2 := S2;
  A2 := Copy( A2, 1, L );
  if (UpperCase( A1 ) = UpperCase( A2 )) and (A1 <> A2) then
  asm
    int 3;
  end;
  Result := StrLEq1( S1, S2, L );
end;}

procedure DoConversion;
const MaxBufLen = 1023;
var Line: Integer;

  procedure Error( const Msg: String );
  begin
    Writeln( 'Error in line ' + Int2Str( Line ) + ': ' + Msg
             //,MB_OK or MB_APPLMODAL or MB_SETFOREGROUND
           );
    Readln;
    Halt;
  end;

var Fin, Fout, F: PStream;
    C, C2: Char;
    Cmd: String;
    Skipping: Boolean;
    ReplaceFrom, ReplaceTo: PStrList;
    SkipFrom, SkipTo: PStrList;
    SkipWait: String;
    Symbols: PStrList;
    SkipCommandsUntilEndIf, IfEndLevel: Integer;
    OK: Boolean;
    Buf: array[ 0..MaxBufLen ] of Char;
    BufPos: Integer;

  procedure AddChar( Ch: Char );
  var I, L, L2: Integer;
  begin
    if SkipWait = '' then
      Fout.Write( Ch, 1 );
    if Ch < ' ' then
    begin
      BufPos := 0;
      Buf[ 0 ] := #0;
    end
      else
    begin
      Buf[ BufPos ] := Ch;
      Inc( BufPos );
      Buf[ BufPos ] := #0;
      if BufPos >= MaxBufLen then
      begin
        Move( Buf[ 100 ], Buf[ 0 ], BufPos - 100 );
        Dec( BufPos, 100 );
      end;
      if SkipWait <> '' then
      begin
        L := Length( SkipWait );
        if L <= BufPos then
        if StrLEq( @ SkipWait[ 1 ], @ Buf[ BufPos - L ], L ) then
        begin
          BufPos := 0;
          Buf[ 0 ] := #0;
          SkipWait := '';
        end;
        Exit;
      end;
      if SkipWait = '' then
      begin
        for I := SkipFrom.Count-1 downto 0 do
        begin
          L := StrLen( SkipFrom.ItemPtrs[ I ] );
          if L <= BufPos then
          if StrLEq( SkipFrom.ItemPtrs[ I ],
                     @ Buf[ BufPos - L ], L ) then
          begin
            BufPos := 0;
            Buf[ 0 ] := #0;
            Fout.Position := Fout.Position - DWORD( L );
            SkipWait := SkipTo.Items[ I ];
            Exit;
          end;
        end;
      end;
      for I := ReplaceFrom.Count-1 downto 0 do
      begin
        L := StrLen( ReplaceFrom.ItemPtrs[ I ] );
        if (L <= BufPos) then
        if StrLEq( ReplaceFrom.ItemPtrs[ I ],
                   @ Buf[ BufPos - L ], L ) then
        begin
          L2 := StrLen( ReplaceTo.ItemPtrs[ I ] );
          Move( ReplaceTo.ItemPtrs[ I ]^, Buf[ 0 ], L2 + 1 );
          BufPos := L2;
          Fout.Position := Fout.Position - DWORD( L );
          Fout.WriteStr( Buf );
          Exit;
        end;
      end;
    end;
  end;

  procedure Waiting( const S: String );
    procedure WaitingChar( W: Char );
    begin
     if Fin.Read( C, 1 ) <> 1 then Exit;
     if C <> W then Error( Cmd + ' waiting for "' + W + '", but "' + C +
        '" found.' );
    end;
  var I: Integer;
  begin
    for I := 1 to Length( S ) do WaitingChar( S[ I ] );
  end;

var S: String;
    I, L: Integer;
    Flag: Boolean;
    IncludeStrm: PStream;
    InputFileName: String;
begin
  Writeln( 'Converts KOL.PAS to FPC-compatible. Key options:' );
  Writeln( '/I:inputfile - path is relative to current, by default /I:KOL.PAS' );
  Writeln( '/O:outputdir - relative to current, by default /O:FPC' );
  Writeln( '/D:symbol    - used in {#If[symbol]}, {#If not[symbol]} commands' );

  S := '';
  for I := 1 to ParamCount do
  begin
    S := ParamStr( I );
    if StrIsStartingFrom( PChar( UpperCase( S ) ), '/I:' ) then
      S := CopyEnd( S, 4 )
    else
    if StrIsStartingFrom( PChar( UpperCase( S ) ), '/I' ) then
      S := CopyEnd( S, 3 )
    else
      S := '';
    if S <> '' then break;
  end;
  if S = '' then S := 'KOL.PAS';
  InputFileName := S;
  if pos( '\', S ) <= 0 then
    S := GetStartDir + S;
  F := NewReadFileStream( S );
  if F.Handle = INVALID_HANDLE_VALUE then
    Error( 'Can not open input file: ' + S );
  Fin := NewMemoryStream;
  Stream2StreamEx( Fin, F, MaxInt );
  Fin.Position := 0;
  F.Free;
  Fout := NewMemoryStream;
  Fout.Size := Fin.Size;
  ReplaceFrom := NewStrList;
  ReplaceTo := NewStrList;
  SkipFrom := NewStrList;
  SkipTo := NewStrList;

  Symbols := NewStrList;
  for I := 1 to ParamCount do
  begin
    S := UpperCase( ParamStr( I ) );
    if StrIsStartingFrom( PChar( UpperCase( S ) ), '/D:' ) then
      S := CopyEnd( S, 4 )
    else
    if StrIsStartingFrom( PChar( UpperCase( S ) ), '/D' ) then
      S := CopyEnd( S, 3 )
    else continue;
    Symbols.Add( S );
  end;
  SkipCommandsUntilEndIf := 0; // do not skip
  IfEndLevel := 0;

  Line := 1;
  Skipping := FALSE;
  OK := FALSE;
  SkipWait := '';
  BufPos := 0;
  Buf[ 0 ] := #0;
  TRY
    while Fin.Position < Fin.Size do
    begin
      if Fin.Read( C, 1 ) <> 1 then Exit;
      if C = #13 then Inc( Line );
      if (C <> '{') or (SkipWait <> '') then
      begin
        if not Skipping then AddChar( C );
      end
      else begin
             if Fin.Read( C2, 1 ) <> 1 then Exit;
             if C2 = #13 then Inc( Line );
             if not (C2 in ['#', '+', '-']) then
             begin
               if not Skipping then
               begin
                 AddChar( C ); AddChar( C2 );
               end;
             end
               else
             begin
               Cmd := C2;
               while C <> '}' do
               begin
                 if Fin.Read( C, 1 ) <> 1 then Exit;
                 if C = #13 then Inc( Line );
                 Cmd := Cmd + C;
               end;
               if StrIsStartingFrom( PChar( UpperCase( Cmd ) ),
                  '#END' ) then
               begin
                 SkipCommandsUntilEndIf := SkipCommandsUntilEndIf shr 1;
                 Dec( IfEndLevel );
                 if IfEndLevel < 0 then
                   Error( '{' + Cmd + ': #EndIf without #If' );
               end
                 else
               if SkipCommandsUntilEndIf and 1 <> 0 then
               begin
                 continue; // skip all commands until #End If!
               end
                 else
               if Cmd = '-}' then
               begin
                 if Skipping then Error( Cmd + ' - already in skipping mode.' );
                 Skipping := TRUE;
               end
                 else
               if Cmd = '+}' then
               begin
                 if not Skipping then Error( Cmd + ' - not in skipping mode.' );
                 Skipping := FALSE;
               end
                 else
               if Cmd = '++}' then
               begin
                 Waiting( '(*' );
                 C := '!';
                 while TRUE do
                 begin
                   C2 := C;
                   if Fin.Read( C, 1 ) <> 1 then Exit;
                   if C = #13 then Inc( Line );
                   if (C2 = '*') and (C = ')') then break;
                   AddChar( C );
                 end;
                 Fout.Position := Fout.Position - 1;
                 Waiting( '{--}' );
               end
                 else
               if StrEq(Cmd, '#L}') or StrEq(Cmd, '#Line}') then
               begin
                 Cmd := '{' + InputFileName + ' Line ' + Int2Str( Line ) + '}';
                 Fout.WriteStr( Cmd );
                 BufPos := 0;
                 Buf[ 0 ] := #0;
               end
                 else
               if StrIsStartingFrom( PChar( UpperCase( Cmd ) ), '#REPLACE[' ) then
               begin
                 Parse( Cmd, '[' );
                 S := Parse( Cmd, ']' ); Cmd := Trim( Cmd );
                 if (Cmd + ' ')[ 1 ] <> '[' then Error(
                    ' Replace to what was waiting for in [ ], but ' + Cmd +
                    ' found.' );
                 for I := ReplaceFrom.Count-1 downto 0 do
                 begin
                    L := StrLen( ReplaceFrom.ItemPtrs[ I ] );
                    if (L = Length( S )) then
                    if StrLEq( ReplaceFrom.ItemPtrs[ I ], @ S[ 1 ], L ) then
                    begin
                      ReplaceFrom.Delete( I );
                      ReplaceTo.Delete( I );
                      break;
                    end;
                 end;
                 ReplaceFrom.Add( S );
                 Parse( Cmd, '[' );
                 S := Parse( Cmd, ']' );
                 if Trim( Cmd ) <> '}' then Error(
                    ' Replace to not finished correctly with '']''.' );
                 ReplaceTo.Add( S );
               end
                 else
               if StrIsStartingFrom( PChar( UpperCase( Cmd ) ),
                  '#NOREPLACE[' ) then
               begin
                 Parse( Cmd, '[' );
                 S := Parse( Cmd, ']' ); Cmd := Trim( Cmd );
                 if Cmd <> '}' then Error(
                    ' NoReplace command has ambigous characters: ' + Cmd );
                 for I := ReplaceFrom.Count-1 downto 0 do
                 begin
                    L := StrLen( ReplaceFrom.ItemPtrs[ I ] );
                    if (L = Length( S )) then
                    if StrLEq( ReplaceFrom.ItemPtrs[ I ], @ S[ 1 ], L ) then
                    begin
                      ReplaceFrom.Delete( I );
                      ReplaceTo.Delete( I );
                      break;
                    end;
                 end;
               end
                 else
               if StrIsStartingFrom( PChar( UpperCase( Cmd ) ),
                  '#SKIPBETWEEN[' ) then
               begin
                 Parse( Cmd, '[' );
                 S := Parse( Cmd, ']' ); Cmd := Trim( Cmd );
                 for I := 0 to SkipFrom.Count-1 do
                 begin
                   L := StrLen( SkipFrom.ItemPtrs[ I ] );
                   if L = Length( S ) then
                   if StrLEq( SkipFrom.ItemPtrs[ I ], @ S[ 1 ], L ) then
                   begin
                     SkipFrom.Delete( I );
                     SkipTo.Delete( I );
                     break;
                   end;
                 end;
                 SkipFrom.Add( S );
                 if Trim( S ) = '' then
                   Error( 'Skip between both strings must not be empty.' );
                 Parse( Cmd, '[' );
                 S := Parse( Cmd, ']' ); Cmd := Trim( Cmd );
                 if Cmd <> '}' then
                   Error( ' Skip between not finished correctly with '']''.' );
                 SkipTo.Add( S );
                 if Trim( S ) = '' then
                   Error( 'Skip between both strings must not be empty.' );
               end
                 else
               if StrIsStartingFrom( PChar( UpperCase( Cmd ) ),
                  '#NOSKIPBETWEEN[' ) then
               begin
                 Parse( Cmd, '[' );
                 S := Parse( Cmd, ']' ); Cmd := Trim( Cmd );
                 if Trim( S ) = '' then
                   Error( 'Cancel Skip between string must not be empty.' );
                 for I := 0 to SkipFrom.Count-1 do
                 begin
                   L := StrLen( SkipFrom.ItemPtrs[ I ] );
                   if L = Length( S ) then
                   if StrLEq( SkipFrom.ItemPtrs[ I ], @ S[ 1 ], L ) then
                   begin
                     SkipFrom.Delete( I );
                     SkipTo.Delete( I );
                     S := '';
                     break;
                   end;
                 end;
                 if S <> '' then
                   Error( 'Cancel Skip Betwen string not found in a list.' );
                 if Cmd <> '}' then
                   Error( ' Cancel Skip Between not finished correctly with '']''.' );
               end
                 else
               if StrIsStartingFrom( PChar( UpperCase( Cmd ) ),
                  '#IF' ) then
               begin
                 S := UpperCase( Trim( CopyEnd( Cmd, 4 ) ) );
                 Flag := StrIsStartingFrom( PChar( S ), 'NOT' );
                 Parse( S, '[' );
                 S := Trim( Parse( S, ']' ) );
                 if S = '' then
                   Error( 'Symbol not defined in #IF command: ' + Cmd );
                 SkipCommandsUntilEndIf := (SkipCommandsUntilEndIf shl 1) or
                                        (SkipCommandsUntilEndIf and 1);
                 if not Flag and (Symbols.IndexOf( S ) < 0) or
                        Flag and (Symbols.IndexOf( S ) >= 0)
                   then SkipCommandsUntilEndIf := SkipCommandsUntilEndIf or 1;
                 Inc( IfEndLevel );
                 if IfEndLevel > 32 then
                   Error( 'More then 32 levels of nested #If commands.' );
               end
                 else
               if StrIsStartingFrom( PChar( UpperCase( Cmd ) ),
                  '#INCLUDE' ) then
               begin
                 Parse( Cmd, '[' );
                 S := Trim( Parse( Cmd, ']' ) );
                 if not FileExists( GetStartDir + S ) then
                   Error( 'File ' + S + ' not found to include it' );
                 BufPos := 0;
                 Buf[ 0 ] := #0;
                 IncludeStrm := NewReadFileStream( GetStartDir + S );
                 Stream2Stream( Fout, IncludeStrm, IncludeStrm.Size );
                 IncludeStrm.Free;
               end
                 else
               if StrEq( Cmd, '#LIST}' ) then
               begin
                 Fout.WriteStr( '(*'#13#10'List of replacements:'#13#10 );
                 for I := 0 to ReplaceFrom.Count-1 do
                   Fout.WriteStr( '[' + ReplaceFrom.Items[ I ] + '][' +
                                  ReplaceTo.Items[ I ] + ']'#13#10 );
                 Fout.WriteStr( #13#10'List of skips:'#13#10 );
                 for I := 0 to SkipFrom.Count-1 do
                   Fout.WriteStr( '[' + SkipFrom.Items[ I ] + '][' +
                                  SkipTo.Items[ I ] + ']'#13#10 );
                 Fout.WriteStr( '*)'#13#10 );
                 BufPos := 0;
                 Buf[ 0 ] := #0;
               end
                 else
               begin
                 Error( '{' + Cmd + '- Unknown command.' );
               end;
             end;
           end;
    end;
    OK := TRUE;
  FINALLY
    Fin.Free;
    if OK then
    begin
      S := '';
      for I := 1 to ParamCount do
      begin
        S := ParamStr( I );
        if StrIsStartingFrom( PChar( UpperCase( S ) ), '/O:' ) then
          S := CopyEnd( S, 4 )
        else
        if StrIsStartingFrom( PChar( UpperCase( S ) ), '/O' ) then
          S := CopyEnd( S, 3 )
        else
          S := '';
        if S <> '' then break;
      end;
      if S = '' then
        S := 'FPC';
      S := IncludeTrailingPathDelimiter( S );
      ForceDirectories( GetStartDir + S );
      S := GetStartDir + S + InputFileName;
      F := NewWriteFileStream( S );
      if F.Handle <> INVALID_HANDLE_VALUE then
      begin
        Fout.Size := Fout.Position;
        Fout.Position := 0;
        if Stream2StreamEx( F, Fout, MaxInt ) = 0 then
          OK := FALSE;
      end
        else OK := FALSE;
      F.Free;
    end;
    Fout.Free;
    ReplaceFrom.Free;
    ReplaceTo.Free;
    SkipFrom.Free;
    SkipTo.Free;
  END;
  if not OK then
    Error( 'Can not write file ' + S + ': ' + SysErrorMessage( GetLastError ) )
  else
    Writeln( InputFileName + ' converted OK.' );
end;

begin
  UseDelphiMemoryManager;
  UseInputOutput;
  DoConversion;
end.
